"""
This example covers a basic polymorphism concept, avoid Data Ambiguity.
"""
from dataclasses import dataclass
from typing import List
from datamodel import BaseModel, Column

@dataclass
class Animal:
    name: str

@dataclass
class Tiger(Animal):
    weight_lbs: int

@dataclass
class Elephant(Animal):
    trunk_length_cm: int

@dataclass
class Zoo:
    animals: List[Animal]

animals = [Elephant(name="Eldor", trunk_length_cm=176), Tiger(name="Roy", weight_lbs=405)]
zoo = Zoo(animals=animals)

print('ZOO DATACLASS: ', zoo)
for animal in zoo.animals:
    print('DC: ', animal)

class Animal(BaseModel):
    name: str
    weight: int

class Snake(Animal):
    length: float

class Mammal(Animal):
    temp: float = Column(default=38.0)

class Tiger(Mammal):
    height: float

class Elephant(Mammal):
    trunk_length: int

class Zoo(BaseModel):
    animals: List[Animal]

animals = [
    Elephant(name="Eldor", trunk_length=176),
    Tiger(name="Roy", weight=405, height=389),
    Snake(name='Robin', length=210)
]
zoo = Zoo(animals=animals)

print('ZOO: ', zoo)
for animal in zoo.animals:
    print(animal)
# print: ZOO: Zoo(animals=[Elephant(name='Eldor', weight=0, temp=38.0, trunk_length=176), Tiger(name='Roy', weight=405, temp=38.0, height=389.0), Snake(name='Robin', weight=0, length=210.0)])



from datetime import (
    datetime,
    time,
    date
)
import uuid
from decimal import Decimal
from datamodel import BaseModel, Column


def auto_now_add(*args, **kwargs):
    return uuid.uuid4()

def is_employee(obj) -> str:
    if obj in ('Y', 'F'):
        return obj
    elif obj is True:
        return 'Y'
    else:
        return 'F'

class Contact(BaseModel):
    account: str = ''
    value: str = ''

class User(BaseModel):
    """
    User Basic Structure
    """
    id: uuid.UUID = Column(primary_key=True, required=True, default=auto_now_add(), db_default='uuid_generate_v4()')
    firstname: str
    lastname: str
    name: str = Column(required=True, default='John Doe')
    age: int = Column(default=18, required=True)
    salary: Decimal = Column(default=10.0)
    in_time: time = Column(default='15:00')
    out_time: time = Column(default='23:00')
    birth: date = Column(required=False)
    is_employee: str = Column(required=True, default='F', encoder=is_employee)
    size: float
    signup_ts: datetime = Column(default=datetime.now(), db_default='now()')
    contacts: Contact = Column(required=False)

    class Meta:
        name = 'users'
        schema = 'public'
        driver = 'pg'
        strict = False

u = User()
print(u.model(dialect='json'))

### creates a new user:
jesus = User(
    firstname='Jesus', lastname='Lara', age=43, salary=1500.25, size=185.28, birth='1978-10-23', in_time='11:00:00.000', out_time='23:59:00.000', is_employee=True
)
print(jesus)
b = jesus.json()
jlara = User.from_json(b)
print(jlara)



import inspect
from datamodel import BaseModel, Field

class HazelPortable(BaseModel):
    factory_id: int = Field(default=1)

    def write_portable(self, writer):
        for name, f in self.columns().items():
            _type = f.type
            value = getattr(self, f)
            print('WRITER ', _type, value)

    def read_portable(self, reader):
        for name, f in self.columns().items():
            _type = f.type
            print('READER', name, _type)

    def set_factory(self, fid: int = 1):
        self.factory_id = fid

    def get_factory_id(self):
        return self.factory_id

    def get_class_id(self):
        return self.factory_id

class Customer(HazelPortable):
    name: str
    age: int
    is_active: bool

print(inspect.getmro(Customer))

c1 = Customer(name='Hola', age=33, is_active=True)
c2 = Customer(name='Hola', age=33, is_active=True)

print('Is Instance: ', isinstance(c1, HazelPortable))
print('Is a Subclass: ', issubclass(Customer, HazelPortable))

print('EQUALITY :: ', c1 == c2)



from datetime import datetime
from typing import List, Optional
_started = datetime.now()
from pydantic import AnyUrl
from pydantic.dataclasses import dataclass
_end = datetime.now()

print(f'Pydantic Loaded at: {_end - _started}')

@dataclass
class Address:
    street: str
    zipcode: int

@dataclass
class User:
    id: int
    name: str = 'John Doe'
    signup_ts: datetime = None
    address: Optional[Address] = None


user = User(id='42', signup_ts='2032-06-21T12:00', address=Address(street="Calle Mayor", zipcode=45510))
print(user)

@dataclass
class NavbarButton:
    href: AnyUrl

@dataclass
class Navbar:
    button: NavbarButton

navbar = Navbar(button=NavbarButton('https://example.com'))
print(navbar)

_end = datetime.now()
print(f'Pydantic End at: {_end - _started}')

_started = datetime.now()
from datamodel import BaseModel
from dataclasses import dataclass

_end = datetime.now()
print(f'BaseModel Loaded at: {_end - _started}')

@dataclass
class Address:
    street: str
    zipcode: int

class User(BaseModel):
    id: int
    name: str = 'John Doe'
    signup_ts: datetime = None
    address: Optional[Address]

user = User(id='42', signup_ts='2032-06-21T12:00', address={"street": "Calle Mayor", "zipcode": 45510})
print(user)

class HttpsUrl(AnyUrl):
    def __init__(self, url, scheme: str = 'https'):
        super(HttpsUrl, self).__init__(
            url=url,
            scheme=scheme
        )

class NavbarButton(BaseModel):
    href: HttpsUrl

class Navbar(BaseModel):
    button: NavbarButton

navbar = Navbar(button='https://example.com')
print(navbar)

_end = datetime.now()
print(f'BaseModel End at: {_end - _started}')



"""
JSON Encoders.
"""
from .json import JSONContent

DefaultEncoder = JSONContent

class BaseEncoder:
    """
    Encoder replacement for json.dumps using orjson
    """

    def __init__(self, *args, **kwargs):
        encoder = DefaultEncoder(*args, **kwargs)
        self.encode = encoder.__call__



import uuid
from datetime import date, datetime, time
from decimal import Decimal

from datamodel import BaseModel, Column, Field
from datamodel.exceptions import ValidationError


def auto_now_add():
    return uuid.uuid4()

def is_employee(obj) -> str:
    if obj in ('Y', 'F'):
        return obj
    elif obj is True:
        return 'Y'
    else:
        return 'F'

class Contact(BaseModel):
    account: str = ''
    value: str = ''

class User(BaseModel):
    """
    User Basic Structure
    """
    id: uuid.UUID = Column(primary_key=True, required=True, default=auto_now_add, db_default='uuid_generate_v4()')
    firstname: str = Field(required=True)
    lastname: str = Field(required=True)
    name: str = Column(required=True, default='John Doe')
    age: int = Column(default=18, required=True)
    salary: Decimal = Column(default=10.0)
    in_time: time = Column(default='15:00')
    out_time: time = Column(default='23:00')
    birth: date = Column(required=False)
    is_employee: str = Column(required=True, default='F', encoder=is_employee)
    size: float
    signup_ts: datetime = Column(default=datetime.now(), db_default='now()')
    contacts: Contact = Column(required=False)

    class Meta:
        strict: bool = True


# Error on empy user:
try:
    u = User(firstname='Jesus', lastname='Lara', birth='Egg')
except ValidationError as ex:
    print(ex)
except TypeError as ex:
    print(ex)
except ValueError as ex:
    print(ex)

### creates a new user:
try:
    jesus = User(
        firstname='Jesus', lastname='Lara', age='Hola', salary=1500.25, size=185.28, birth='1978-10-23', in_time='11:00:00.000', out_time='23:59:00.000', is_employee=True
    )
    print(jesus)
except ValidationError as ex:
    print(ex)
    print(ex.payload)

# try to adding a missing keyword:
try:
    jesus = User(airport=True)
except ValidationError as ex:
    print(ex)
except TypeError as ex:
    print(ex)

class Animal(BaseModel):
    name: str
    specie: str
    age: int

try:
    animal = Animal(**{
        "name": "Human",
        "specie": "Homo Sapiens",
        "age": "otra verga"
    })
except ValidationError as ex:
    print(ex)
    print(ex.payload)
except TypeError as ex:
    print(ex)
except ValueError as ex:
    print(ex)



#!/usr/bin/env python
"""DataModels.

    Dataclass Reimplementation with true inheritance (without decorators.)
See:
https://github.com/phenobarbital/DataModel
"""
import ast
from os import path

from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup


def get_path(filename):
    return path.join(path.dirname(path.abspath(__file__)), filename)


def readme():
    with open(get_path('README.md'), encoding='utf-8') as rd:
        return rd.read()


version = get_path('datamodel/version.py')
with open(version, 'r', encoding='utf-8') as meta:
    # exec(meta.read())
    t = compile(meta.read(), version, 'exec', ast.PyCF_ONLY_AST)
    for node in (n for n in t.body if isinstance(n, ast.Assign)):
        if len(node.targets) == 1:
            name = node.targets[0]
            if isinstance(name, ast.Name) and \
                    name.id in (
                            '__version__',
                            '__title__',
                            '__description__',
                            '__author__',
                            '__license__', '__author_email__'):
                        v = node.value
                        if name.id == '__version__':
                            __version__ = v.s
                        if name.id == '__title__':
                            __title__ = v.s
                        if name.id == '__description__':
                            __description__ = v.s
                        if name.id == '__license__':
                            __license__ = v.s
                        if name.id == '__author__':
                            __author__ = v.s
                        if name.id == '__author_email__':
                            __author_email__ = v.s


COMPILE_ARGS = ["-O2"]

extensions = [
    Extension(
        name='datamodel.fields',
        sources=['datamodel/fields.pyx'],
        extra_compile_args=COMPILE_ARGS,
        language="c"
    ),
    Extension(
        name='datamodel.converters',
        sources=['datamodel/converters.pyx'],
        extra_compile_args=COMPILE_ARGS,
        language="c"
    ),
    Extension(
        name='datamodel.validation',
        sources=['datamodel/validation.pyx'],
        extra_compile_args=COMPILE_ARGS,
        language="c++"
    ),
    Extension(
        name='datamodel.exceptions',
        sources=['datamodel/exceptions.pyx'],
        extra_compile_args=COMPILE_ARGS,
        language="c"
    ),
    Extension(
        name='datamodel.types',
        sources=['datamodel/types.pyx'],
        extra_compile_args=COMPILE_ARGS,
        language="c"
    ),
    Extension(
        name='datamodel.parsers.json',
        sources=['datamodel/parsers/json.pyx'],
        extra_compile_args=COMPILE_ARGS,
        language="c++"
    )
]

setup(
    name="python-datamodel",
    version=__version__,
    python_requires=">=3.8.1",
    url="https://github.com/phenobarbital/python-datamodel",
    description=__description__,
    keywords=['asyncio', 'dataclass', 'dataclasses', 'data models'],
    platforms=['any'],
    long_description=readme(),
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Framework :: AsyncIO",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: System :: Systems Administration",
        "Topic :: Utilities",
        "Environment :: Web Environment",
    ],
    author=__author__,
    author_email=__author_email__,
    packages=find_packages(exclude=["contrib", "docs", "tests"]),
    include_package_data=True,
    license=__license__,
    setup_requires=[
        "setuptools==65.4.1",
        "wheel==0.37.1",
        "Cython==0.29.32",
        "asyncio==3.4.3",
        "cchardet==2.1.7"
    ],
    install_requires=[
        "wheel==0.37.1",
        "Cython==0.29.32",
        "numpy==1.23.4",
        "uvloop==0.17.0",
        "asyncio==3.4.3",
        "cchardet==2.1.7",
        "objectpath==0.6.1",
        "orjson==3.8.0",
        'typing_extensions==4.4.0',
        "asyncpg==0.26.0",
        "python-dateutil==2.8.2"
    ],
    tests_require=[
        'pytest>=6.0.0',
        'pytest-asyncio==0.19.0',
        'pytest-xdist==2.1.0',
        'pytest-assume==2.4.3'
    ],
    test_suite='tests',
    ext_modules=cythonize(extensions),
    project_urls={  # Optional
        "Source": "https://github.com/phenobarbital/datamodels",
        "Funding": "https://paypal.me/phenobarbital",
        "Say Thanks!": "https://saythanks.io/to/phenobarbital",
    },
)
